# Copyright 2020 The Trieste Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This module is the home of the sampling functionality required by Trieste's
acquisition functions.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Callable, Generic

import tensorflow as tf
import tensorflow_probability as tfp
from scipy.optimize import bisect

from ..models import ProbabilisticModel
from ..models.interfaces import HasTrajectorySampler, ProbabilisticModelType, SupportsPredictY
from ..types import TensorType
from .utils import select_nth_output


class ThompsonSampler(ABC, Generic[ProbabilisticModelType]):
    r"""
    A :class:`ThompsonSampler` samples either the minimum values or minimisers of a function
    modeled by an underlying :class:`ProbabilisticModel` across a  discrete set of points.
    """

    def __init__(self, sample_min_value: bool = False):
        """
        :sample_min_value: If True then sample from the minimum value of the function,
            else sample the function's minimiser.
        """
        self._sample_min_value = sample_min_value

    @property
    def sample_min_value(self) -> bool:
        """Whether this samples from the minimum value of the function
        (as opposed to the function's minimiser)."""
        return self._sample_min_value

    def __repr__(self) -> str:
        """"""
        return f"""{self.__class__.__name__}(
        {self._sample_min_value})
        """

    @abstractmethod
    def sample(
        self,
        model: ProbabilisticModelType,
        sample_size: int,
        at: TensorType,
        select_output: Callable[[TensorType], TensorType] = select_nth_output,
    ) -> TensorType:
        """
        :param model: The model to sample from.
        :param sample_size: The desired number of samples.
        :param at: Input points that define the sampler.
        :param select_output: A method that returns the desired output from the model sampler, with
            shape `[S, N]` where `S` is the number of samples and `N` is the number of locations.
            Defaults to the :func:~`trieste.acquisition.utils.select_nth_output` function with
            output dimension 0.
        :return: Samples.
        """


class ExactThompsonSampler(ThompsonSampler[ProbabilisticModel]):
    r"""
    This sampler provides exact Thompson samples of the objective function's
    minimiser :math:`x^*` over a discrete set of input locations.
    Although exact Thompson sampling is costly (incuring with an :math:`O(N^3)` complexity to
    sample over a set of `N` locations), this method can be used for any probabilistic model
    with a sampling method.
    """

    def sample(
        self,
        model: ProbabilisticModel,
        sample_size: int,
        at: TensorType,
        select_output: Callable[[TensorType], TensorType] = select_nth_output,
    ) -> TensorType:
        """
        Return exact samples from either the objective function's minimiser or its minimal value
        over the candidate set `at`. Note that minimiser ties aren't broken randomly.

        :param model: The model to sample from.
        :param sample_size: The desired number of samples.
        :param at: Where to sample the predictive distribution, with shape `[N, D]`, for points
            of dimension `D`.
        :param select_output: A method that returns the desired output from the model sampler, with
            shape `[S, N]` where `S` is the number of samples and `N` is the number of locations.
            Defaults to the :func:~`trieste.acquisition.utils.select_nth_output` function with
            output dimension 0.
        :return: The samples, of shape `[S, D]` (where `S` is the `sample_size`) if sampling
            the function's minimiser or shape `[S, 1]` if sampling the function's mimimal value.
        :raise ValueError: If ``at`` has an invalid shape or if ``sample_size`` is not positive.
        """
        tf.debugging.assert_positive(sample_size)
        tf.debugging.assert_shapes([(at, ["N", None])])

        samples = select_output(model.sample(at, sample_size))[..., None]  # [S, N, 1]

        if self._sample_min_value:
            thompson_samples = tf.reduce_min(samples, axis=1)  # [S, 1]
        else:
            samples_2d = tf.squeeze(samples, -1)  # [S, N]
            indices = tf.math.argmin(samples_2d, axis=1)
            thompson_samples = tf.gather(at, indices)  # [S, D]

        return thompson_samples


class GumbelSampler(ThompsonSampler[ProbabilisticModel]):
    r"""
    This sampler follows :cite:`wang2017max` and yields approximate samples of the objective
    minimum value :math:`y^*` via the empirical cdf :math:`\operatorname{Pr}(y^*<y)`. The cdf
    is approximated by a Gumbel distribution
    .. math:: \mathcal G(y; a, b) = 1 - e^{-e^\frac{y - a}{b}}
    where :math:`a, b \in \mathbb R` are chosen such that the quartiles of the Gumbel and cdf match.
    Samples are obtained via the Gumbel distribution by sampling :math:`r` uniformly from
    :math:`[0, 1]` and applying the inverse probability integral transform
    :math:`y = \mathcal G^{-1}(r; a, b)`.
    Note that the :class:`GumbelSampler` can only sample a function's minimal value and not
    its minimiser.
    """

    def __init__(self, sample_min_value: bool = False):
        """
        :sample_min_value: If True then sample from the minimum value of the function,
            else sample the function's minimiser.
        """

        if not sample_min_value:
            raise ValueError(
                f"""
                Gumbel samplers can only sample a function's minimal value,
                however received sample_min_value={sample_min_value}
                """
            )

        super().__init__(sample_min_value)

    def sample(
        self,
        model: ProbabilisticModel,
        sample_size: int,
        at: TensorType,
        select_output: Callable[[TensorType], TensorType] = select_nth_output,
    ) -> TensorType:
        """
        Return approximate samples from of the objective function's minimum value.

        :param model: The model to sample from.
        :param sample_size: The desired number of samples.
        :param at: Points at where to fit the Gumbel distribution, with shape `[N, D]`, for points
            of dimension `D`. We recommend scaling `N` with search space dimension.
        :param select_output: A method that returns the desired output from the model sampler, with
            shape `[S, N]` where `S` is the number of samples and `N` is the number of locations.
            Currently unused.
        :return: The samples, of shape `[S, 1]`, where `S` is the `sample_size`.
        :raise ValueError: If ``at`` has an invalid shape or if ``sample_size`` is not positive.
        """
        tf.debugging.assert_positive(sample_size)
        tf.debugging.assert_shapes([(at, ["N", None])])

        if isinstance(model, SupportsPredictY):
            fmean, fvar = model.predict_y(at)
        else:
            fmean, fvar = model.predict(at)

        fsd = tf.math.sqrt(fvar)

        def probf(y: tf.Tensor) -> tf.Tensor:  # Build empirical CDF for Pr(y*^hat<y)
            unit_normal = tfp.distributions.Normal(
                tf.constant(0, fmean.dtype), tf.constant(1, fmean.dtype)
            )
            log_cdf = unit_normal.log_cdf(-(y - fmean) / fsd)
            return 1 - tf.exp(tf.reduce_sum(log_cdf, axis=0))

        left = tf.reduce_min(fmean - 5 * fsd)
        right = tf.reduce_max(fmean + 5 * fsd)

        def binary_search(val: float) -> float:  # Find empirical interquartile range
            return bisect(lambda y: probf(y) - val, left, right, maxiter=10000)

        q1, q2 = map(binary_search, [0.25, 0.75])

        log = tf.math.log
        l1 = log(log(4.0 / 3.0))
        l2 = log(log(4.0))
        b = (q1 - q2) / (l1 - l2)
        a = (q2 * l1 - q1 * l2) / (l1 - l2)

        uniform_samples = tf.random.uniform([sample_size], dtype=fmean.dtype)
        gumbel_samples = log(-log(1 - uniform_samples)) * tf.cast(b, fmean.dtype) + tf.cast(
            a, fmean.dtype
        )
        gumbel_samples = tf.expand_dims(gumbel_samples, axis=-1)  # [S, 1]
        return gumbel_samples


class ThompsonSamplerFromTrajectory(ThompsonSampler[HasTrajectorySampler]):
    r"""
    This sampler provides approximate Thompson samples of the objective function's
    minimiser :math:`x^*` by minimizing approximate trajectories sampled from the
    underlying probabilistic model. This sampling method can be used for any
    probabilistic model with a :meth:`trajectory_sampler` method.
    """

    def sample(
        self,
        model: ProbabilisticModel,
        sample_size: int,
        at: TensorType,
        select_output: Callable[[TensorType], TensorType] = select_nth_output,
    ) -> TensorType:
        """
        Return approximate samples from either the objective function's minimser or its minimal
        value over the candidate set `at`. Note that minimiser ties aren't broken randomly.

        :param model: The model to sample from.
        :param sample_size: The desired number of samples.
        :param at: Where to sample the predictive distribution, with shape `[N, D]`, for points
            of dimension `D`.
        :param select_output: A method that returns the desired output from the model sampler, with
            shape `[S, N]` where `S` is the number of samples and `N` is the number of locations.
            Defaults to the :func:~`trieste.acquisition.utils.select_nth_output` function with
            output dimension 0.
        :return: The samples, of shape `[S, D]` (where `S` is the `sample_size`) if sampling
            the function's minimser or shape `[S, 1]` if sampling the function's mimimal value.
        :raise ValueError: If ``at`` has an invalid shape or if ``sample_size`` is not positive.
        """
        tf.debugging.assert_positive(sample_size)
        tf.debugging.assert_shapes([(at, ["N", None])])

        if not isinstance(model, HasTrajectorySampler):
            raise ValueError(
                f"Thompson sampling from trajectory only supports models with a trajectory_sampler "
                f"method; received {model!r}"
            )

        trajectory_sampler = model.trajectory_sampler()

        if self._sample_min_value:
            thompson_samples = tf.zeros([0, 1], dtype=at.dtype)  # [0,1]
        else:
            thompson_samples = tf.zeros([0, tf.shape(at)[1]], dtype=at.dtype)  # [0,D]

        for _ in tf.range(sample_size):
            sampled_trajectory = trajectory_sampler.get_trajectory()
            expanded_at = tf.expand_dims(at, -2)  # [N, 1, D]
            evaluated_trajectory = select_output(sampled_trajectory(expanded_at))  # [N, 1]
            if self._sample_min_value:
                sample = tf.reduce_min(evaluated_trajectory, keepdims=True)  # [1, 1]
            else:
                sample = tf.gather(at, tf.math.argmin(evaluated_trajectory))  # [1, D]

            thompson_samples = tf.concat([thompson_samples, sample], axis=0)

        return thompson_samples  # [S, D] or [S, 1]



# Copyright 2021 The Trieste Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
""" This folder contains multi-objective optimization utilities. """
from .dominance import non_dominated
from .pareto import Pareto, get_reference_point
from .partition import (
    DividedAndConquerNonDominated,
    ExactPartition2dNonDominated,
    prepare_default_non_dominated_partition_bounds,
)



# Copyright 2021 The Trieste Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This package contains the primary interface for deep neural network models. It also contains a
number of :class:`TrainableProbabilisticModel` wrappers for neural network models. Note that
currently copying/saving models is not supported, so when
:class:`~trieste.bayesian_optimizer.BayesianOptimizer` is used ``track_state`` should be set
to `False`.
"""

from .architectures import GaussianNetwork, KerasEnsemble, KerasEnsembleNetwork
from .builders import build_keras_ensemble
from .interface import DeepEnsembleModel, KerasPredictor
from .models import DeepEnsemble
from .sampler import DeepEnsembleTrajectorySampler, deep_ensemble_trajectory
from .utils import get_tensor_spec_from_data, negative_log_likelihood, sample_with_replacement



# Copyright 2021 The Trieste Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Callable, Generic, Optional, Sequence, TypeVar, overload

import gpflow
import tensorflow as tf
from check_shapes import check_shapes, inherit_check_shapes
from typing_extensions import Protocol, final, runtime_checkable

from ..data import Dataset
from ..space import EncoderFunction
from ..types import TensorType
from ..utils import DEFAULTS

ProbabilisticModelType = TypeVar(
    "ProbabilisticModelType", bound="ProbabilisticModel", contravariant=True
)
""" Contravariant type variable bound to :class:`~trieste.models.ProbabilisticModel`.
This is used to specify classes such as samplers and acquisition function builders that
take models as input parameters and might ony support models with certain features. """


@runtime_checkable
class ProbabilisticModel(Protocol):
    """A probabilistic model.

    NOTE: This and its subclasses are defined as Protocols rather than ABCs in order to allow
    acquisition functions to depend on the intersection of different model types. As a result, it
    is also possible to pass models to acquisition functions that don't explicitly inherit from
    this class, as long as they implement all the necessary methods. This may change in future if
    https://github.com/python/typing/issues/213 is implemented.
    """

    @abstractmethod
    @check_shapes(
        "query_points: [batch..., D]",
        "return[0]: [batch..., E...]",
        "return[1]: [batch..., E...]",
    )
    def predict(self, query_points: TensorType) -> tuple[TensorType, TensorType]:
        """
        Return the mean and variance of the independent marginal distributions at each point in
        ``query_points``.

        This is essentially a convenience method for :meth:`predict_joint`, where non-event
        dimensions of ``query_points`` are all interpreted as broadcasting dimensions instead of
        batch dimensions, and the covariance is squeezed to remove redundant nesting.

        :param query_points: The points at which to make predictions, of shape [..., D].
        :return: The mean and variance of the independent marginal distributions at each point in
            ``query_points``. For a predictive distribution with event shape E, the mean and
            variance will both have shape [...] + E.
        """
        raise NotImplementedError

    @abstractmethod
    @check_shapes(
        "query_points: [batch..., N, D]",
        "return: [batch..., S, N, E...]",
    )
    def sample(self, query_points: TensorType, num_samples: int) -> TensorType:
        """
        Return ``num_samples`` samples from the independent marginal distributions at
        ``query_points``.

        :param query_points: The points at which to sample, with shape [..., N, D].
        :param num_samples: The number of samples at each point.
        :return: The samples. For a predictive distribution with event shape E, this has shape
            [..., S, N] + E, where S is the number of samples.
        """
        raise NotImplementedError

    @abstractmethod
    def log(self, dataset: Optional[Dataset] = None) -> None:
        """
        Log model-specific information at a given optimization step.

        :param dataset: Optional data that can be used to log additional data-based model summaries.
        """
        raise NotImplementedError


@runtime_checkable
class TrainableProbabilisticModel(ProbabilisticModel, Protocol):
    """A trainable probabilistic model."""

    @abstractmethod
    def update(self, dataset: Dataset) -> None:
        """
        Update the model given the specified ``dataset``. Does not train the model.

        :param dataset: The data with which to update the model.
        """
        raise NotImplementedError

    @abstractmethod
    def optimize(self, dataset: Dataset) -> Any:
        """
        Optimize the model objective with respect to (hyper)parameters given the specified
        ``dataset``.

        :param dataset: The data with which to train the model.
        :return: Any (optimizer-specific) optimization result.
        """
        raise NotImplementedError


@runtime_checkable
class SupportsPredictJoint(ProbabilisticModel, Protocol):
    """A probabilistic model that supports predict_joint."""

    @abstractmethod
    @check_shapes(
        "query_points: [batch..., B, D]",
        "return[0]: [batch..., B, E...]",
        "return[1]: [batch..., E..., B, B]",
    )
    def predict_joint(self, query_points: TensorType) -> tuple[TensorType, TensorType]:
        """
        :param query_points: The points at which to make predictions, of shape [..., B, D].
        :return: The mean and covariance of the joint marginal distribution at each batch of points
            in ``query_points``. For a predictive distribution with event shape E, the mean will
            have shape [..., B] + E, and the covariance shape [...] + E + [B, B].
        """
        raise NotImplementedError


@runtime_checkable
class SupportsPredictY(ProbabilisticModel, Protocol):
    """A probabilistic model that supports predict_y."""

    @abstractmethod
    @check_shapes(
        "query_points: [broadcast batch..., D]",
        "return[0]: [batch..., E...]",
        "return[1]: [batch..., E...]",
    )
    def predict_y(self, query_points: TensorType) -> tuple[TensorType, TensorType]:
        """
        Return the mean and variance of the independent marginal distributions at each point in
        ``query_points`` for the observations, including noise contributions.

        :param query_points: The points at which to make predictions, of shape [..., D].
        :return: The mean and variance of the independent marginal distributions at each point in
            ``query_points``. For a predictive distribution with event shape E, the mean and
            variance will both have shape [...] + E.
        """
        raise NotImplementedError


@runtime_checkable
class SupportsGetKernel(ProbabilisticModel, Protocol):
    """A probabilistic model that supports get_kernel."""

    @abstractmethod
    def get_kernel(self) -> gpflow.kernels.Kernel:
        """
        Return the kernel of the model.
        :return: The kernel.
        """
        raise NotImplementedError


@runtime_checkable
class TrainableSupportsGetKernel(TrainableProbabilisticModel, SupportsGetKernel, Protocol):
    """A trainable probabilistic model that supports get_kernel."""


@runtime_checkable
class SupportsGetObservationNoise(ProbabilisticModel, Protocol):
    """A probabilistic model that supports get_observation_noise."""

    @abstractmethod
    def get_observation_noise(self) -> TensorType:
        """
        Return the variance of observation noise.

        :return: The observation noise.
        """
        raise NotImplementedError


@runtime_checkable
class SupportsGetInternalData(ProbabilisticModel, Protocol):
    """A probabilistic model that stores and has access to its own training data."""

    @abstractmethod
    def get_internal_data(self) -> Dataset:
        """
        Return the model's training data.

        :return: The model's training data.
        """
        raise NotImplementedError


@runtime_checkable
class SupportsGetMeanFunction(ProbabilisticModel, Protocol):
    """A probabilistic model that makes use of a mean function."""

    @abstractmethod
    def get_mean_function(self) -> Callable[[TensorType], TensorType]:
        """
        Return the model's mean function, i.e. a parameterized function that can explain
        coarse scale variations in the data, leaving just the residuals to be explained by
        our model.

        :return: The model's mean function.
        """
        raise NotImplementedError


@runtime_checkable
class FastUpdateModel(ProbabilisticModel, Protocol):
    """A model with the ability to predict based on (possibly fantasized) supplementary data."""

    @abstractmethod
    def conditional_predict_f(
        self, query_points: TensorType, additional_data: Dataset
    ) -> tuple[TensorType, TensorType]:
        """
        Return the mean and variance of the independent marginal distributions at each point in
        ``query_points``, given an additional batch of (possibly fantasized) data.

        :param query_points: The points at which to make predictions, of shape [M, D].
        :param additional_data: Dataset with query_points with shape [..., N, D] and observations
                 with shape [..., N, L]
        :return: The mean and variance of the independent marginal distributions at each point in
            ``query_points``, with shape [..., L, M, M].
        """
        raise NotImplementedError

    @abstractmethod
    def conditional_predict_joint(
        self, query_points: TensorType, additional_data: Dataset
    ) -> tuple[TensorType, TensorType]:
        """
        :param query_points: The points at which to make predictions, of shape [M, D].
        :param additional_data: Dataset with query_points with shape [..., N, D] and observations
                 with shape [..., N, L]
        :return: The mean and covariance of the joint marginal distribution at each batch of points
            in ``query_points``, with shape [..., L, M, M].
        """
        raise NotImplementedError

    @abstractmethod
    def conditional_predict_f_sample(
        self, query_points: TensorType, additional_data: Dataset, num_samples: int
    ) -> TensorType:
        """
        Return ``num_samples`` samples from the independent marginal distributions at
        ``query_points``, given an additional batch of (possibly fantasized) data.

        :param query_points: The points at which to sample, with shape [..., N, D].
        :param additional_data: Dataset with query_points with shape [..., N, D] and observations
                 with shape [..., N, L]
        :param num_samples: The number of samples at each point.
        :return: The samples. For a predictive distribution with event shape E, this has shape
            [..., S, N] + E, where S is the number of samples.
        """
        raise NotImplementedError

    def conditional_predict_y(
        self, query_points: TensorType, additional_data: Dataset
    ) -> tuple[TensorType, TensorType]:
        """
        Return the mean and variance of the independent marginal distributions at each point in
        ``query_points`` for the observations, including noise contributions, given an additional
        batch of (possibly fantasized) data.

        Note that this is not supported by all models.

        :param query_points: The points at which to make predictions, of shape [M, D].
        :param additional_data: Dataset with query_points with shape [..., N, D] and observations
                 with shape [..., N, L]
        :return: The mean and variance of the independent marginal distributions at each point in
            ``query_points``.
        """
        raise NotImplementedError(
            f"Model {self!r} does not support predicting observations, just the latent function"
        )


@runtime_checkable
class HasTrajectorySampler(ProbabilisticModel, Protocol):
    """A probabilistic model that has an associated trajectory sampler."""

    def trajectory_sampler(
        self: ProbabilisticModelType,
    ) -> TrajectorySampler[ProbabilisticModelType]:
        """
        Return a trajectory sampler that supports this model.

        :return: The trajectory sampler.
        """
        raise NotImplementedError


@runtime_checkable
class HasReparamSampler(ProbabilisticModel, Protocol):
    """A probabilistic model that has an associated reparametrization sampler."""

    def reparam_sampler(
        self: ProbabilisticModelType, num_samples: int
    ) -> ReparametrizationSampler[ProbabilisticModelType]:
        """
        Return a reparametrization sampler providing `num_samples` samples.

        :param num_samples: The desired number of samples.
        :return: The reparametrization sampler.
        """
        raise NotImplementedError


@runtime_checkable
class SupportsReparamSamplerObservationNoise(
    HasReparamSampler, SupportsGetObservationNoise, Protocol
):
    """A model that supports both reparam_sampler and get_observation_noise."""


class ModelStack(ProbabilisticModel, Generic[ProbabilisticModelType]):
    r"""
    A :class:`ModelStack` is a wrapper around a number of :class:`ProbabilisticModel`\ s of type
    :class:`ProbabilisticModelType`. It combines the outputs of each model for predictions and
    sampling.

    **Note:** Only supports vector outputs (i.e. with event shape [E]). Outputs for any two models
    are assumed independent. Each model may itself be single- or multi-output, and any one
    multi-output model may have dependence between its outputs. When we speak of *event size* in
    this class, we mean the output dimension for a given :class:`ProbabilisticModel`,
    whether that is the :class:`ModelStack` itself, or one of the subsidiary
    :class:`ProbabilisticModel`\ s within the :class:`ModelStack`. Of course, the event
    size for a :class:`ModelStack` will be the sum of the event sizes of each subsidiary model.
    """

    def __init__(
        self,
        model_with_event_size: tuple[ProbabilisticModelType, int],
        *models_with_event_sizes: tuple[ProbabilisticModelType, int],
    ):
        r"""
        The order of individual models specified at :meth:`__init__` determines the order of the
        :class:`ModelStack` output dimensions.

        :param model_with_event_size: The first model, and the size of its output events.
            **Note:** This is a separate parameter to ``models_with_event_sizes`` simply so that the
            method signature requires at least one model. It is not treated specially.
        :param \*models_with_event_sizes: The other models, and sizes of their output events.
        """
        self._models, self._event_sizes = zip(*(model_with_event_size,) + models_with_event_sizes)

    # NB we don't use @inherit_shapes below as some classes break the shape API (👀 fantasizer)
    # instead we rely on the shape checking inside the submodels

    def predict(self, query_points: TensorType) -> tuple[TensorType, TensorType]:
        r"""
        :param query_points: The points at which to make predictions, of shape [..., D].
        :return: The predictions from all the wrapped models, concatenated along the event axis in
            the same order as they appear in :meth:`__init__`. If the wrapped models have predictive
            distributions with event shapes [:math:`E_i`], the mean and variance will both have
            shape [..., :math:`\sum_i E_i`].
        """
        means, vars_ = zip(*[model.predict(query_points) for model in self._models])
        return tf.concat(means, axis=-1), tf.concat(vars_, axis=-1)

    def sample(self, query_points: TensorType, num_samples: int) -> TensorType:
        r"""
        :param query_points: The points at which to sample, with shape [..., N, D].
        :param num_samples: The number of samples at each point.
        :return: The samples from all the wrapped models, concatenated along the event axis. For
            wrapped models with predictive distributions with event shapes [:math:`E_i`], this has
            shape [..., S, N, :math:`\sum_i E_i`], where S is the number of samples.
        """
        samples = [model.sample(query_points, num_samples) for model in self._models]
        return tf.concat(samples, axis=-1)

    def log(self, dataset: Optional[Dataset] = None) -> None:
        """
        Log model-specific information at a given optimization step.

        :param dataset: Optional data that can be used to log additional data-based model summaries.
        """
        for i, model in enumerate(self._models):
            with tf.name_scope(f"{i}"):
                model.log(dataset)


class TrainableModelStack(ModelStack[TrainableProbabilisticModel], TrainableProbabilisticModel):
    r"""
    A :class:`TrainableModelStack` is a wrapper around a number of
    :class:`TrainableProbabilisticModel`\ s.
    It delegates training data to each model for updates and optimization.

    :class:`TrainableProbabilisticModel`\ s within the :class:`TrainableModelStack`.
    Of course, the event size for a :class:`TrainableModelStack` will be the sum of the
    event sizes of each subsidiary model.
    """

    def update(self, dataset: Dataset) -> None:
        """
        Update all the wrapped models on their corresponding data. The data for each model is
        extracted by splitting the observations in ``dataset`` along the event axis according to the
        event sizes specified at :meth:`__init__`.

        :param dataset: The query points and observations for *all* the wrapped models.
        """
        observations = tf.split(dataset.observations, self._event_sizes, axis=-1)

        for model, obs in zip(self._models, observations):
            model.update(Dataset(dataset.query_points, obs))

    def optimize(self, dataset: Dataset) -> Sequence[Any]:
        """
        Optimize all the wrapped models on their corresponding data. The data for each model is
        extracted by splitting the observations in ``dataset`` along the event axis according to the
        event sizes specified at :meth:`__init__`.

        :param dataset: The query points and observations for *all* the wrapped models.
        """
        observations = tf.split(dataset.observations, self._event_sizes, axis=-1)
        results = []

        for model, obs in zip(self._models, observations):
            results.append(model.optimize(Dataset(dataset.query_points, obs)))

        return results


class HasReparamSamplerModelStack(ModelStack[HasReparamSampler], HasReparamSampler):
    r"""
    A :class:`PredictJointModelStack` is a wrapper around a number of
    :class:`HasReparamSampler`\ s.
    It provides a  :meth:`reparam_sampler` method only if all the submodel samplers
    are the same.
    """

    def reparam_sampler(self, num_samples: int) -> ReparametrizationSampler[HasReparamSampler]:
        """
        Return a reparameterization sampler providing `num_samples` samples across
        all the models in the model stack. This is currently only implemented for
        stacks made from models that have a :class:`BatchReparametrizationSampler`
        as their reparameterization sampler.

        :param num_samples: The desired number of samples.
        :return: The reparametrization sampler.
        :raise NotImplementedError: If the models in the stack do not share the
            same :meth:`reparam_sampler`.
        """

        samplers = [model.reparam_sampler(num_samples) for model in self._models]
        unique_sampler_types = set(type(sampler) for sampler in samplers)
        if len(unique_sampler_types) == 1:
            # currently assume that all sampler constructors look the same
            shared_sampler_type = type(samplers[0])
            return shared_sampler_type(num_samples, self)
        else:
            raise NotImplementedError(
                f"""
                Reparameterization sampling is only currently supported for model
                stacks built from models that use the same reparameterization sampler,
                however, received samplers of types {unique_sampler_types}.
                """
            )


class PredictJointModelStack(ModelStack[SupportsPredictJoint], SupportsPredictJoint):
    r"""
    A :class:`PredictJointModelStack` is a wrapper around a number of
    :class:`SupportsPredictJoint`\ s.
    It delegates :meth:`predict_joint` to each model.
    """

    def predict_joint(self, query_points: TensorType) -> tuple[TensorType, TensorType]:
        r"""
        :param query_points: The points at which to make predictions, of shape [..., B, D].
        :return: The predictions from all the wrapped models, concatenated along the event axis in
            the same order as they appear in :meth:`__init__`. If the wrapped models have predictive
            distributions with event shapes [:math:`E_i`], the mean will have shape
            [..., B, :math:`\sum_i E_i`], and the covariance shape
            [..., :math:`\sum_i E_i`, B, B].
        """
        means, covs = zip(*[model.predict_joint(query_points) for model in self._models])
        return tf.concat(means, axis=-1), tf.concat(covs, axis=-3)


class PredictYModelStack(ModelStack[SupportsPredictY], SupportsPredictY):
    r"""
    A :class:`PredictYModelStack` is a wrapper around a number of
    :class:`SupportsPredictY`\ s.
    It delegates :meth:`predict_y` to each model.
    """

    def predict_y(self, query_points: TensorType) -> tuple[TensorType, TensorType]:
        r"""
        :param query_points: The points at which to make predictions, of shape [..., D].
        :return: The predictions from all the wrapped models, concatenated along the event axis in
            the same order as they appear in :meth:`__init__`. If the wrapped models have predictive
            distributions with event shapes [:math:`E_i`], the mean and variance will both have
            shape [..., :math:`\sum_i E_i`].
        :raise NotImplementedError: If any of the models don't implement predict_y.
        """
        means, vars_ = zip(*[model.predict_y(query_points) for model in self._models])
        return tf.concat(means, axis=-1), tf.concat(vars_, axis=-1)


# It's useful, though a bit ugly, to define the stack constructors for some model type combinations
class TrainableSupportsPredictJoint(TrainableProbabilisticModel, SupportsPredictJoint, Protocol):
    """A model that is both trainable and supports predict_joint."""


class TrainablePredictJointModelStack(
    TrainableModelStack, PredictJointModelStack, ModelStack[TrainableSupportsPredictJoint]
):
    """A stack of models that are both trainable and support predict_joint."""


class TrainableSupportsPredictY(TrainableProbabilisticModel, SupportsPredictY, Protocol):
    """A model that is both trainable and supports predict_y."""


class TrainablePredictYModelStack(
    TrainableModelStack, PredictYModelStack, ModelStack[TrainableSupportsPredictY]
):
    """A stack of models that are both trainable and support predict_y."""


class SupportsPredictJointPredictY(SupportsPredictJoint, SupportsPredictY, Protocol):
    """A model that supports both predict_joint and predict_y."""


class PredictJointPredictYModelStack(
    PredictJointModelStack, PredictYModelStack, ModelStack[SupportsPredictJointPredictY]
):
    """A stack of models that support both predict_joint and predict_y."""


class TrainableSupportsPredictJointHasReparamSampler(
    TrainableSupportsPredictJoint, HasReparamSampler, Protocol
):
    """A model that is trainable, supports predict_joint and has a reparameterization sampler."""


class TrainablePredictJointReparamModelStack(
    TrainablePredictJointModelStack,
    HasReparamSamplerModelStack,
    ModelStack[TrainableSupportsPredictJointHasReparamSampler],
):
    """A stack of models that are both trainable and support predict_joint."""


class ReparametrizationSampler(ABC, Generic[ProbabilisticModelType]):
    r"""
    These samplers employ the *reparameterization trick* to draw samples from a
    :class:`ProbabilisticModel`\ 's predictive distribution  across a discrete set of
    points. See :cite:`wilson2018maximizing` for details.
    """

    def __init__(self, sample_size: int, model: ProbabilisticModelType):
        r"""
        Note that our :class:`TrainableModelStack` currently assumes that
        all :class:`ReparametrizationSampler` constructors have **only** these inputs
        and so will not work with more complicated constructors.

        :param sample_size: The desired number of samples.
        :param model: The model to sample from.
        :raise ValueError (or InvalidArgumentError): If ``sample_size`` is not positive.
        """
        tf.debugging.assert_positive(sample_size)

        self._sample_size = sample_size
        self._model = model
        self._initialized = tf.Variable(False)  # Keep track of when we need to resample

    def __repr__(self) -> str:
        """"""
        return f"{self.__class__.__name__}({self._sample_size!r}, {self._model!r})"

    @abstractmethod
    def sample(self, at: TensorType, *, jitter: float = DEFAULTS.JITTER) -> TensorType:
        """
        :param at: Where to sample the predictive distribution, with shape `[..., 1, D]`, for points
            of dimension `D`.
        :param jitter: The size of the jitter to use when stabilising the Cholesky decomposition of
            the covariance matrix.
        :return: The samples, of shape `[..., S, B, L]`, where `S` is the `sample_size`, `B` is
            the number of points per batch, and `L` is the number of latent model dimensions.
        """

        raise NotImplementedError

    def reset_sampler(self) -> None:
        """
        Reset the sampler so that new samples are drawn at the next :meth:`sample` call.
        """
        self._initialized.assign(False)


TrajectoryFunction = Callable[[TensorType], TensorType]
"""
Type alias for trajectory functions. These have similar behaviour to an :const:`AcquisitionFunction`
but have additional sampling properties and support multiple model outputs.

An :const:`TrajectoryFunction` evaluates a batch of `B` samples, each across different sets
of `N` query points (of dimension `D`) i.e. takes input of shape `[N, B, D]` and returns
shape `[N, B, L]`, where `L` is the number of outputs of the model. Note that we require the `L`
dimension to be present, even if there is only one output.

A key property of these trajectory functions is that the same sample draw is evaluated
for all queries. This property is known as consistency.
"""


class TrajectoryFunctionClass(ABC):
    """
    An :class:`TrajectoryFunctionClass` is a trajectory function represented using a class
    rather than as a standalone function. Using a class to represent a trajectory function
    makes it easier to update and resample without having to retrace the function.
    """

    @abstractmethod
    def __call__(self, x: TensorType) -> TensorType:
        """Call trajectory function."""


class TrajectorySampler(ABC, Generic[ProbabilisticModelType]):
    r"""
    This class builds functions that approximate a trajectory sampled from an
    underlying :class:`ProbabilisticModel`.

    Unlike the :class:`ReparametrizationSampler`, a :class:`TrajectorySampler` provides
    consistent samples (i.e ensuring that the same sample draw is used for all evaluations
    of a particular trajectory function).
    """

    def __init__(self, model: ProbabilisticModelType):
        """
        :param model: The model to sample from.
        """
        self._model = model

    def __repr__(self) -> str:
        """"""
        return f"{self.__class__.__name__}({self._model!r})"

    @abstractmethod
    def get_trajectory(self) -> TrajectoryFunction:
        """
        Sample a batch of `B` trajectories. Note that the batch size `B` is determined
        by the first call of the :const:`TrajectoryFunction`. To change the batch size
        of a :const:`TrajectoryFunction` after initialization, you must
        recall :meth:`get_trajectory`.

        :return: A trajectory function representing an approximate trajectory
            from the model, taking an input of shape `[N, B, D]` and returning shape `[N, B, L]`,
            where `L` is the number of outputs of the model.
        """
        raise NotImplementedError

    def resample_trajectory(self, trajectory: TrajectoryFunction) -> TrajectoryFunction:
        """
        A :const:`TrajectoryFunction` can often be efficiently updated in-place to provide
        a new sample without retracing. Note that if the underlying :class:`ProbabilisticModel`
        has been updated, then we must call :meth:`update_trajectory` to get a new sample from
        the new model.

        Efficient implementations of a :class:`TrajectorySampler` will have a custom method here
        to allow in-place resampling. However, the default behavior is just to make a new
        trajectory from scratch.

        :param trajectory: The trajectory function to be resampled.
        :return: The new resampled trajectory function.
        """
        return self.get_trajectory()

    def update_trajectory(self, trajectory: TrajectoryFunction) -> TrajectoryFunction:
        """
        Update a :const:`TrajectoryFunction` to reflect an update in its
        underlying :class:`ProbabilisticModel` and resample accordingly.

        Efficient implementations will have a custom method here to allow in-place resampling
        and updating. However, the default behavior is just to make a new trajectory from scratch.

        :param trajectory: The trajectory function to be resampled.
        :return: The new trajectory function updated for a new model
        """
        return self.get_trajectory()


@runtime_checkable
class SupportsGetInducingVariables(ProbabilisticModel, Protocol):
    """A probabilistic model uses and has access to an inducing point approximation."""

    @abstractmethod
    def get_inducing_variables(self) -> tuple[TensorType, TensorType, TensorType, bool]:
        """
        Return the model's inducing variables.

        :return: Tensors containing: the inducing points (i.e. locations of the inducing
            variables); the variational mean q_mu; the Cholesky decomposition of the
            variational covariance q_sqrt; and a bool denoting if we are using whitened
            or not whitened representations.
        """
        raise NotImplementedError


@runtime_checkable
class SupportsCovarianceWithTopFidelity(ProbabilisticModel, Protocol):
    """A probabilistic model is multifidelity and has access to a method to calculate the
    covariance between a point and the same point at the top fidelity"""

    @property
    @abstractmethod
    def num_fidelities(self) -> int:
        """
        The number of fidelities
        """
        raise NotImplementedError

    @abstractmethod
    def covariance_with_top_fidelity(self, query_points: TensorType) -> TensorType:
        """
        Calculate the covariance of the output at `query_point` and a given fidelity with the
        highest fidelity output at the same `query_point`.

        :param query_points: The query points to calculate the covariance for, of shape [N, D+1],
            where the final column of the final dimension contains the fidelity of the query point
        :return: The covariance with the top fidelity for the `query_points`, of shape [N, P]
        """
        raise NotImplementedError


def encode_dataset(dataset: Dataset, encoder: EncoderFunction) -> Dataset:
    """Return a new Dataset with the query points encoded using the given encoder."""
    return Dataset(encoder(dataset.query_points), dataset.observations)


class EncodedProbabilisticModel(ProbabilisticModel):
    """A probabilistic model with an associated query point encoder.

    Classes that inherit from this (or the other associated mixins below) should implement the
    relevant _encoded methods (e.g. predict_encoded instead of predict), to which the public
    methods delegate after encoding their input. Take care to use the correct methods internally
    to avoid encoding twice accidentally.
    """

    @property
    @abstractmethod
    def encoder(self) -> EncoderFunction | None:
        """Query point encoder."""

    @overload
    def encode(self, points: TensorType) -> TensorType: ...

    @overload
    def encode(self, points: Dataset) -> Dataset: ...

    def encode(self, points: Dataset | TensorType) -> Dataset | TensorType:
        """Encode points or a Dataset using the query point encoder."""
        if self.encoder is None:
            return points
        elif isinstance(points, Dataset):
            return encode_dataset(points, self.encoder)
        else:
            return self.encoder(points)

    @abstractmethod
    def predict_encoded(self, query_points: TensorType) -> tuple[TensorType, TensorType]:
        """Implementation of predict on encoded query points."""

    @abstractmethod
    def sample_encoded(self, query_points: TensorType, num_samples: int) -> TensorType:
        """Implementation of sample on encoded query points."""

    @final
    @inherit_check_shapes
    def predict(self, query_points: TensorType) -> tuple[TensorType, TensorType]:
        return self.predict_encoded(self.encode(query_points))

    @final
    @inherit_check_shapes
    def sample(self, query_points: TensorType, num_samples: int) -> TensorType:
        return self.sample_encoded(self.encode(query_points), num_samples)


class EncodedTrainableProbabilisticModel(EncodedProbabilisticModel, TrainableProbabilisticModel):
    """A trainable probabilistic model with an associated query point encoder."""

    @abstractmethod
    def update_encoded(self, dataset: Dataset) -> None:
        """Implementation of update on the encoded dataset."""

    @abstractmethod
    def optimize_encoded(self, dataset: Dataset) -> Any:
        """Implementation of optimize on the encoded dataset."""

    @final
    def update(self, dataset: Dataset) -> None:
        return self.update_encoded(self.encode(dataset))

    @final
    def optimize(self, dataset: Dataset) -> Any:
        return self.optimize_encoded(self.encode(dataset))


class EncodedSupportsPredictJoint(EncodedProbabilisticModel, SupportsPredictJoint):
    """A probabilistic model that supports predict_joint with an associated query point encoder."""

    @abstractmethod
    def predict_joint_encoded(self, query_points: TensorType) -> tuple[TensorType, TensorType]:
        """Implementation of predict_joint on encoded query points."""

    @final
    @inherit_check_shapes
    def predict_joint(self, query_points: TensorType) -> tuple[TensorType, TensorType]:
        return self.predict_joint_encoded(self.encode(query_points))


class EncodedSupportsPredictY(EncodedProbabilisticModel, SupportsPredictY):
    """A probabilistic model that supports predict_y with an associated query point encoder."""

    @abstractmethod
    def predict_y_encoded(self, query_points: TensorType) -> tuple[TensorType, TensorType]:
        """Implementation of predict_y on encoded query points."""

    @final
    @inherit_check_shapes
    def predict_y(self, query_points: TensorType) -> tuple[TensorType, TensorType]:
        return self.predict_y_encoded(self.encode(query_points))


class EncodedFastUpdateModel(EncodedProbabilisticModel, FastUpdateModel):
    """A fast update model with an associated query point encoder."""

    @abstractmethod
    def conditional_predict_f_encoded(
        self, query_points: TensorType, additional_data: Dataset
    ) -> tuple[TensorType, TensorType]:
        """Implementation of conditional_predict_f on encoded query points."""

    @abstractmethod
    def conditional_predict_joint_encoded(
        self, query_points: TensorType, additional_data: Dataset
    ) -> tuple[TensorType, TensorType]:
        """Implementation of conditional_predict_joint on encoded query points."""

    @abstractmethod
    def conditional_predict_f_sample_encoded(
        self, query_points: TensorType, additional_data: Dataset, num_samples: int
    ) -> TensorType:
        """Implementation of conditional_predict_f_sample on encoded query points."""

    @abstractmethod
    def conditional_predict_y_encoded(
        self, query_points: TensorType, additional_data: Dataset
    ) -> tuple[TensorType, TensorType]:
        """Implementation of conditional_predict_y on encoded query points."""

    @final
    def conditional_predict_f(
        self, query_points: TensorType, additional_data: Dataset
    ) -> tuple[TensorType, TensorType]:
        return self.conditional_predict_f_encoded(
            self.encode(query_points), self.encode(additional_data)
        )

    @final
    def conditional_predict_joint(
        self, query_points: TensorType, additional_data: Dataset
    ) -> tuple[TensorType, TensorType]:
        return self.conditional_predict_joint_encoded(
            self.encode(query_points), self.encode(additional_data)
        )

    @final
    def conditional_predict_f_sample(
        self, query_points: TensorType, additional_data: Dataset, num_samples: int
    ) -> TensorType:
        return self.conditional_predict_f_sample_encoded(
            self.encode(query_points), self.encode(additional_data), num_samples
        )

    @final
    def conditional_predict_y(
        self, query_points: TensorType, additional_data: Dataset
    ) -> tuple[TensorType, TensorType]:
        return self.conditional_predict_y_encoded(
            self.encode(query_points), self.encode(additional_data)
        )


def get_encoder(model: ProbabilisticModel) -> EncoderFunction | None:
    """Helper function for getting an encoder from model (which may or may not have one)."""
    if isinstance(model, EncodedProbabilisticModel):
        return model.encoder
    return None



# Copyright 2021 The Trieste Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This file contains implementations of neural network architectures with Keras.
"""

from __future__ import annotations

import contextlib
from abc import abstractmethod
from typing import Any, Callable, Sequence

import dill
import numpy as np
import tensorflow as tf
import tensorflow_probability as tfp
from gpflow.keras import tf_keras

try:
    SafeModeScope = tf_keras.src.saving.serialization_lib.SafeModeScope
except AttributeError:  # pragma: no cover (tested but not by coverage)
    try:
        # in TF 2.14 it's in keras but not in tf_keras (but vice versa in TF 2.16)!
        from keras.src.saving.serialization_lib import SafeModeScope  # type:ignore[no-redef]
    except (ImportError, ModuleNotFoundError):
        SafeModeScope = contextlib.nullcontext
from tensorflow_probability.python.layers.distribution_layer import DistributionLambda, _serialize

from trieste.types import TensorType


class KerasEnsemble:
    """
    This class builds an ensemble of neural networks, using Keras. Individual networks must
    be instance of :class:`~trieste.models.keras_networks.KerasEnsembleNetwork`. This class
    is meant to be used with :class:`~trieste.models.keras_networks.DeepEnsemble` model wrapper,
    which compiles the model.
    """

    def __init__(
        self,
        networks: Sequence[KerasEnsembleNetwork],
    ) -> None:
        """
        :param networks: A list of neural network specifications, one for each member of the
            ensemble. The ensemble will be built using these specifications.
        :raise ValueError: If there are no objects in ``networks`` or we try to create
            a model with networks whose input or output shapes are not the same.
        """
        if not networks:
            raise ValueError(
                f"networks should consist of KerasEnsembleNetwork objects, however"
                f"received {networks} instead."
            )

        input_shapes, output_shapes, output_dtypes = [], [], []
        for index, network in enumerate(networks):
            network.network_name = f"model_{index}_"
            input_shapes.append(network.input_tensor_spec.shape)
            output_shapes.append(network.output_tensor_spec.shape)
            output_dtypes.append(network.output_tensor_spec.dtype)

        if not all(x == input_shapes[0] for x in input_shapes):
            raise ValueError(
                f"Input shapes for all networks must be the same, however"
                f"received {input_shapes} instead."
            )
        if not all(x == output_shapes[0] for x in output_shapes):
            raise ValueError(
                f"Output shapes for all networks must be the same, however"
                f"received {output_shapes} instead."
            )
        if not all(x == output_dtypes[0] for x in output_dtypes):
            raise ValueError(
                f"Output dtypes for all networks must be the same, however"
                f"received {output_dtypes} instead."
            )
        self.num_outputs = networks[0].flattened_output_shape
        self.output_dtype = networks[0].output_tensor_spec.dtype

        self._networks = networks

        self._model = self._build_ensemble()

    def __repr__(self) -> str:
        """"""
        return f"KerasEnsemble({self._networks!r})"

    @property
    def model(self) -> tf_keras.Model:
        """Returns built but uncompiled Keras ensemble model."""
        return self._model

    @property
    def ensemble_size(self) -> int:
        """
        Returns the size of the ensemble, that is, the number of base learners or individual neural
        network models in the ensemble.
        """
        return len(self._networks)

    def _build_ensemble(self) -> tf_keras.Model:
        """
        Builds the ensemble model by combining all the individual networks in a single Keras model.
        This method relies on ``connect_layers`` method of :class:`KerasEnsembleNetwork` objects
        to construct individual networks.

        :return: The Keras model.
        """
        inputs, outputs = zip(*[network.connect_layers() for network in self._networks])

        return tf_keras.Model(inputs=inputs, outputs=outputs)

    def __getstate__(self) -> dict[str, Any]:
        # When pickling use to_json to save the model.
        state = self.__dict__.copy()
        state["_model"] = self._model.to_json()
        state["_weights"] = self._model.get_weights()

        # Save the history callback (serializing any model)
        if self._model.history:
            history_model = self._model.history.model
            try:
                if history_model is self._model:
                    # no need to serialize the main model, just use a special value instead
                    self._model.history.model = ...
                elif history_model:
                    self._model.history.model = (
                        history_model.to_json(),
                        history_model.get_weights(),
                    )
                state["_history"] = dill.dumps(self._model.history)
            finally:
                self._model.history.model = history_model

        return state

    def __setstate__(self, state: dict[str, Any]) -> None:
        # When unpickling restore the model using model_from_json.
        self.__dict__.update(state)
        # TF 2.15 disallows loading lambdas without "safe-mode" being disabled
        # unfortunately, tfp.layers.DistributionLambda uses lambdas
        with SafeModeScope(False):
            self._model = tf_keras.models.model_from_json(
                state["_model"], custom_objects={"MultivariateNormalTriL": MultivariateNormalTriL}
            )
        self._model.set_weights(state["_weights"])

        # Restore the history (including any model it contains)
        if "_history" in state:
            self._model.history = dill.loads(state["_history"])
            if self._model.history.model is ...:
                self._model.history.set_model(self._model)
            elif self._model.history.model:
                model_json, weights = self._model.history.model
                model = tf_keras.models.model_from_json(
                    model_json,
                    custom_objects={"MultivariateNormalTriL": MultivariateNormalTriL},
                )
                model.set_weights(weights)
                self._model.history.set_model(model)


class KerasEnsembleNetwork:
    """
    This class is an interface that defines necessary attributes and methods for neural networks
    that are meant to be used for building ensembles by
    :class:`~trieste.models.keras_networks.KerasEnsemble`. Subclasses are not meant to
    build and compile Keras models, instead they are providing specification that
    :class:`~trieste.models.keras_networks.KerasEnsemble` will use to build the Keras model.
    """

    def __init__(
        self,
        input_tensor_spec: tf.TensorSpec,
        output_tensor_spec: tf.TensorSpec,
        network_name: str = "",
    ):
        """
        :param input_tensor_spec: Tensor specification for the input to the network.
        :param output_tensor_spec: Tensor specification for the output of the network.
        :param network_name: The name to be used when building the network.
        """
        if not isinstance(input_tensor_spec, tf.TensorSpec):
            raise ValueError(
                f"input_tensor_spec must be an instance of tf.TensorSpec, "
                f"received {type(input_tensor_spec)} instead."
            )
        if not isinstance(output_tensor_spec, tf.TensorSpec):
            raise ValueError(
                f"output_tensor_spec must be an instance of tf.TensorSpec, "
                f"received {type(output_tensor_spec)} instead."
            )

        self.input_tensor_spec = input_tensor_spec
        self.output_tensor_spec = output_tensor_spec
        self.network_name = network_name

    @property
    def input_layer_name(self) -> str:
        return self.network_name + "input"

    @property
    def output_layer_name(self) -> str:
        return self.network_name + "output"

    @property
    def flattened_output_shape(self) -> int:
        return int(np.prod(self.output_tensor_spec.shape))

    @abstractmethod
    def connect_layers(self) -> tuple[tf.Tensor, tf.Tensor]:
        """
        Connects the layers of the neural network. Architecture, layers and layer specifications
        need to be defined by the subclasses.

        :return: Input and output tensor of the network, required by :class:`tf.keras.Model` to
            build a model.
        """
        raise NotImplementedError


class MultivariateNormalTriL(tfp.layers.MultivariateNormalTriL):  # type: ignore[misc]
    """Fixed version of tfp.layers.MultivariateNormalTriL that handles saving."""

    def __init__(
        self,
        event_size: int,
        convert_to_tensor_fn: Callable[
            [tfp.python.distributions.Distribution], TensorType
        ] = tfp.python.distributions.Distribution.sample,
        validate_args: bool = False,
        **kwargs: Any,
    ) -> None:
        self._event_size = event_size
        self._validate_args = validate_args
        super().__init__(event_size, convert_to_tensor_fn, validate_args, **kwargs)

    def get_config(self) -> dict[str, Any]:
        config = {
            "event_size": self._event_size,
            "validate_args": self._validate_args,
            "convert_to_tensor_fn": _serialize(self._convert_to_tensor_fn),
        }
        # skip DistributionLambda's get_config because we don't want to serialize the
        # make_distribution_fn: both to avoid confusing the constructor, and because it doesn't
        # seem to work in TF2.4.
        base_config = super(DistributionLambda, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))


class GaussianNetwork(KerasEnsembleNetwork):
    """
    This class defines layers of a probabilistic neural network using Keras. The network
    architecture is a multilayer fully-connected feed-forward network, with Gaussian
    distribution as an output. The layers are meant to be built as an ensemble model by
    :class:`KerasEnsemble`. Note that this is not a Bayesian neural network.
    """

    def __init__(
        self,
        input_tensor_spec: tf.TensorSpec,
        output_tensor_spec: tf.TensorSpec,
        hidden_layer_args: Sequence[dict[str, Any]] = (
            {"units": 50, "activation": "relu"},
            {"units": 50, "activation": "relu"},
        ),
        independent: bool = False,
    ):
        """
        :param input_tensor_spec: Tensor specification for the input to the network.
        :param output_tensor_spec: Tensor specification for the output of the network.
        :param hidden_layer_args: Specification for building dense hidden layers. Each element in
            the sequence should be a dictionary containing arguments (keys) and their values for a
            :class:`~tf.keras.layers.Dense` hidden layer. Please check Keras Dense layer API for
            available arguments. Objects in the sequence will sequentially be used to add
            :class:`~tf.keras.layers.Dense` layers. Length of this sequence determines the number of
            hidden layers in the network. Default value is two hidden layers, 50 nodes each, with
            ReLu activation functions. Empty sequence needs to be passed to have no hidden layers.
        :param independent: In case multiple outputs are modeled, if set to `True` then
            :class:`~tfp.layers.IndependentNormal` layer
            is used as the output layer. This models outputs as independent, only the diagonal
            elements of the covariance matrix are parametrized. If left as the default `False`,
            then :class:`~tfp.layers.MultivariateNormalTriL` layer is used where correlations
            between outputs are learned as well.
        :raise ValueError: If objects in ``hidden_layer_args`` are not dictionaries.
        """
        super().__init__(input_tensor_spec, output_tensor_spec)

        self._hidden_layer_args = hidden_layer_args
        self._independent = independent

    def _gen_input_tensor(self) -> tf_keras.Input:
        input_tensor = tf_keras.Input(
            shape=self.input_tensor_spec.shape,
            dtype=self.input_tensor_spec.dtype,
            name=self.input_layer_name,
        )
        return input_tensor

    def _gen_hidden_layers(self, input_tensor: tf.Tensor) -> tf.Tensor:
        for index, hidden_layer_args in enumerate(self._hidden_layer_args):
            layer_name = f"{self.network_name}dense_{index}"
            layer = tf_keras.layers.Dense(
                **hidden_layer_args, name=layer_name, dtype=input_tensor.dtype.name
            )
            input_tensor = layer(input_tensor)

        return input_tensor

    def _gen_multi_output_layer(self, input_tensor: tf.Tensor) -> tf.Tensor:
        dist_layer = tfp.layers.IndependentNormal if self._independent else MultivariateNormalTriL
        n_params = dist_layer.params_size(self.flattened_output_shape)

        parameter_layer = tf_keras.layers.Dense(
            n_params, name=self.network_name + "dense_parameters", dtype=input_tensor.dtype.name
        )(input_tensor)

        distribution = dist_layer(
            self.flattened_output_shape,
            tfp.python.distributions.Distribution.mean,
            name=self.output_layer_name,
            dtype=input_tensor.dtype.name,
        )(parameter_layer)

        return distribution

    def _gen_single_output_layer(self, input_tensor: tf.Tensor) -> tf.Tensor:
        parameter_layer = tf_keras.layers.Dense(
            2, name=self.network_name + "dense_parameters", dtype=input_tensor.dtype.name
        )(input_tensor)

        def distribution_fn(inputs: TensorType) -> tfp.distributions.Distribution:
            return tfp.distributions.Normal(inputs[..., :1], tf.math.softplus(inputs[..., 1:]))

        distribution = tfp.layers.DistributionLambda(
            make_distribution_fn=distribution_fn,
            convert_to_tensor_fn=tfp.distributions.Distribution.mean,
            name=self.output_layer_name,
            dtype=input_tensor.dtype.name,
        )(parameter_layer)

        return distribution

    def connect_layers(self) -> tuple[tf.Tensor, tf.Tensor]:
        """
        Connect all layers in the network. We start by generating an input tensor based on input
        tensor specification. Next we generate a sequence of hidden dense layers based on
        hidden layer arguments. Finally, we generate a dense layer whose nodes act as parameters of
        a Gaussian distribution in the final probabilistic layer.

        :return: Input and output tensor of the sequence of layers.
        """
        input_tensor = self._gen_input_tensor()
        hidden_tensor = self._gen_hidden_layers(input_tensor)

        if self.flattened_output_shape == 1:
            output_tensor = self._gen_single_output_layer(hidden_tensor)
        else:
            output_tensor = self._gen_multi_output_layer(hidden_tensor)

        return input_tensor, output_tensor



# Copyright 2021 The Trieste Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
""" This folder contains single-objective optimization functions. """
from .active_learning import (
    BayesianActiveLearningByDisagreement,
    ExpectedFeasibility,
    IntegratedVarianceReduction,
    PredictiveVariance,
    bayesian_active_learning_by_disagreement,
    bichon_ranjan_criterion,
    integrated_variance_reduction,
    predictive_variance,
)
from .continuous_thompson_sampling import (
    GreedyContinuousThompsonSampling,
    ParallelContinuousThompsonSampling,
)
from .entropy import (
    GIBBON,
    MinValueEntropySearch,
    gibbon_quality_term,
    gibbon_repulsion_term,
    min_value_entropy_search,
)
from .function import (
    AugmentedExpectedImprovement,
    BatchExpectedImprovement,
    BatchMonteCarloExpectedImprovement,
    ExpectedConstrainedImprovement,
    ExpectedImprovement,
    FastConstraintsFeasibility,
    MakePositive,
    MonteCarloAugmentedExpectedImprovement,
    MonteCarloExpectedImprovement,
    MultipleOptimismNegativeLowerConfidenceBound,
    NegativeLowerConfidenceBound,
    NegativePredictiveMean,
    ProbabilityOfFeasibility,
    ProbabilityOfImprovement,
    augmented_expected_improvement,
    batch_expected_improvement,
    expected_improvement,
    fast_constraints_feasibility,
    lower_confidence_bound,
    multiple_optimism_lower_confidence_bound,
    probability_below_threshold,
)
from .greedy_batch import Fantasizer, LocalPenalization, hard_local_penalizer, soft_local_penalizer
from .multi_objective import (
    HIPPO,
    BatchMonteCarloExpectedHypervolumeImprovement,
    ExpectedConstrainedHypervolumeImprovement,
    ExpectedHypervolumeImprovement,
    batch_ehvi,
    expected_hv_improvement,
)
from .utils import MultivariateNormalCDF



# Copyright 2020 The Trieste Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from typing import Callable, Optional

import numpy as np
import plotly.graph_objects as go
import tensorflow as tf
from plotly.subplots import make_subplots

from trieste.models.interfaces import ProbabilisticModel
from trieste.types import TensorType
from trieste.utils import to_numpy

from .plotting import create_grid


def format_point_markers(
    num_pts: int,
    num_init: int,
    idx_best: Optional[int] = None,
    mask_fail: Optional[TensorType] = None,
    m_init: str = "x",
    m_add: str = "circle",
    c_pass: str = "green",
    c_fail: str = "red",
    c_best: str = "darkmagenta",
) -> tuple[TensorType, TensorType]:
    """
    Prepares point marker styles according to some BO factors

    :param num_pts: total number of BO points
    :param num_init: initial number of BO points
    :param idx_best: index of the best BO point
    :param mask_fail: Bool vector, True if the corresponding observation violates the constraint(s)
    :param m_init: marker for the initial BO points
    :param m_add: marker for the other BO points
    :param c_pass: color for the regular BO points
    :param c_fail: color for the failed BO points
    :param c_best: color for the best BO points
    :return: 2 string vectors col_pts, mark_pts containing marker styles and colors
    """

    col_pts = np.repeat(c_pass, num_pts).astype("<U15")
    mark_pts = np.repeat(m_init, num_pts).astype("<U15")
    mark_pts[num_init:] = m_add
    if mask_fail is not None:
        col_pts[mask_fail] = c_fail
    if idx_best is not None:
        col_pts[idx_best] = c_best

    return col_pts, mark_pts


def add_surface_plotly(
    xx: TensorType,
    yy: TensorType,
    f: TensorType,
    fig: go.Figure,
    alpha: float = 1.0,
    figrow: int = 1,
    figcol: int = 1,
) -> go.Figure:
    """
    Adds a surface to an existing plotly subfigure

    :param xx: [n, n] array (input)
    :param yy: [n, n] array (input)
    :param f: [n, n] array (output)
    :param fig: the current plotly figure
    :param alpha: transparency
    :param figrow: row index of the subfigure
    :param figcol: column index of the subfigure
    :return: updated plotly figure
    """

    z = f.reshape([xx.shape[0], yy.shape[1]])

    fig.add_trace(
        go.Surface(z=z, x=xx, y=yy, showscale=False, opacity=alpha, colorscale="viridis"),
        row=figrow,
        col=figcol,
    )
    return fig


def add_bo_points_plotly(
    x: TensorType,
    y: TensorType,
    z: TensorType,
    fig: go.Figure,
    num_init: int,
    idx_best: Optional[int] = None,
    mask_fail: Optional[TensorType] = None,
    figrow: int = 1,
    figcol: int = 1,
) -> go.Figure:
    """
    Adds scatter points to an existing subfigure. Markers and colors are chosen according to
    BO factors.
    :param x: [N] x inputs
    :param y: [N] y inputs
    :param z: [N] z outputs
    :param fig: the current plotly figure
    :param num_init: initial number of BO points
    :param idx_best: index of the best BO point
    :param mask_fail: Bool vector, True if the corresponding observation violates the constraint(s)
    :param figrow: row index of the subfigure
    :param figcol: column index of the subfigure
    :return: a plotly figure
    """
    num_pts = x.shape[0]

    col_pts, mark_pts = format_point_markers(num_pts, num_init, idx_best, mask_fail)

    fig.add_trace(
        go.Scatter3d(
            x=x,
            y=y,
            z=z,
            mode="markers",
            marker=dict(size=4, color=col_pts, symbol=mark_pts, opacity=0.8),
        ),
        row=figrow,
        col=figcol,
    )

    return fig


def plot_model_predictions_plotly(
    model: ProbabilisticModel,
    mins: TensorType,
    maxs: TensorType,
    grid_density: int = 20,
    num_samples: Optional[int] = None,
    alpha: float = 0.85,
) -> go.Figure:
    """
    Plots 2-dimensional plot of model's predictions. We first create a regular grid of points
    and evaluate the model on these points. We then plot the mean and 2 standard deviations to
    show epistemic uncertainty.

    For ``DeepGaussianProcess`` models ``num_samples`` should be used
    and set to some positive number. This is needed as predictions from deep GP's are stochastic
    and we need to take more than one sample to estimate the mean and variance.

    :param model: A probabilistic model
    :param mins: List of 2 lower bounds for creating a grid of points for model predictions.
    :param maxs: List of 2 upper bounds for creating a grid of points for model predictions.
    :param grid_density: Number of points per dimension. This will result in a grid size of
        grid_density^2.
    :param num_samples: Number of samples to use with deep GPs.
    :param alpha: Transparency.
    :return: A plotly figure.
    """
    mins = to_numpy(mins)
    maxs = to_numpy(maxs)

    # Create a regular grid on the parameter space
    Xplot, xx, yy = create_grid(mins=mins, maxs=maxs, grid_density=grid_density)

    # Evaluate objective function, ``num_samples`` is currently used
    if num_samples is None:
        Fmean, Fvar = model.predict(Xplot)
    else:
        means = []
        vars = []
        for _ in range(num_samples):
            Fmean_sample, Fvar_sample = model.predict(Xplot)
            means.append(Fmean_sample)
            vars.append(Fvar_sample)
        Fmean = tf.reduce_mean(tf.stack(means), axis=0)
        Fvar = tf.reduce_mean(tf.stack(vars) + tf.stack(means) ** 2, axis=0) - Fmean**2

    n_output = Fmean.shape[1]

    fig = make_subplots(rows=1, cols=n_output, specs=[[{"type": "surface"}] * n_output])

    for k in range(n_output):
        fmean = Fmean[:, k].numpy()
        fvar = Fvar[:, k].numpy()

        lcb = fmean - 2 * np.sqrt(fvar)
        ucb = fmean + 2 * np.sqrt(fvar)

        fig = add_surface_plotly(xx, yy, fmean, fig, alpha=alpha, figrow=1, figcol=k + 1)
        fig = add_surface_plotly(xx, yy, lcb, fig, alpha=alpha - 0.35, figrow=1, figcol=k + 1)
        fig = add_surface_plotly(xx, yy, ucb, fig, alpha=alpha - 0.35, figrow=1, figcol=k + 1)

    fig.update_layout(height=600, width=600)

    return fig


def plot_function_plotly(
    obj_func: Callable[[TensorType], TensorType],
    mins: TensorType,
    maxs: TensorType,
    grid_density: int = 20,
    title: Optional[str] = None,
    xlabel: Optional[str] = None,
    ylabel: Optional[str] = None,
    alpha: float = 1.0,
) -> go.Figure:
    """
    Plots 2-dimensional plot of an objective function. To illustrate the function we create a
    regular grid of points and evaluate the function on these points.

    :param obj_func: The vectorized objective function.
    :param mins: List of 2 lower bounds for creating a grid of points for model predictions.
    :param maxs: List of 2 upper bounds for creating a grid of points for model predictions.
    :param grid_density: Number of points per dimension. This will result in a grid size of
        grid_density^2.
    :param title: optional titles
    :param xlabel: optional xlabel
    :param ylabel: optional ylabel
    :param alpha: transparency
    :return: A plotly figure.
    """

    # Create a regular grid on the parameter space
    Xplot, xx, yy = create_grid(mins=mins, maxs=maxs, grid_density=grid_density)

    # Evaluate objective function
    F = to_numpy(obj_func(Xplot))
    if len(F.shape) == 1:
        F = F.reshape(-1, 1)
    n_output = F.shape[1]

    fig = make_subplots(
        rows=1,
        cols=n_output,
        specs=[[{"type": "surface"}] * n_output],
        subplot_titles=title,
    )

    for k in range(n_output):
        f = F[:, k]
        fig = add_surface_plotly(xx, yy, f, fig, alpha=alpha, figrow=1, figcol=k + 1)
        fig.update_xaxes(title_text=xlabel, row=1, col=k + 1)
        fig.update_yaxes(title_text=ylabel, row=1, col=k + 1)

    fig.update_layout(height=600, width=600)

    return fig
